# -*- coding: utf-8 -*-
"""
Clients used to access third party systems.
"""

from __future__ import absolute_import, unicode_literals

import os
from datetime import datetime
from functools import wraps
from six.moves import urllib

from edx_rest_api_client.client import EdxRestApiClient


class EdxOAuth2APIClient(object):
    """
    Base API Client for accessing edX IDA API endpoints.
    """

    LMS_ROOT_URL = os.environ.get('LMS_ROOT_URL')
    LMS_OAUTH_HOST = os.environ.get('LMS_OAUTH_HOST')
    API_BASE_URL = LMS_ROOT_URL + '/api/'
    APPEND_SLASH = False

    DEFAULT_VALUE_SAFEGUARD = object()

    def __init__(self, client_id=None, client_secret=None):
        """
        Connect to the REST API.
        """
        self.client_id = client_id or os.environ.get('LMS_OAUTH_KEY')
        self.client_secret = client_secret or os.environ.get('LMS_OAUTH_SECRET')
        self.expires_at = datetime.utcnow()
        self.client = None

    def connect(self):
        """
        Connect to the REST API, authenticating with an access token retrieved with our client credentials.
        """
        access_token, expires_at = EdxRestApiClient.get_oauth_access_token(
            self.LMS_OAUTH_HOST + '/oauth2/access_token',
            self.client_id,
            self.client_secret,
            'jwt'
        )
        self.client = EdxRestApiClient(
            self.API_BASE_URL, append_slash=self.APPEND_SLASH, jwt=access_token,
        )
        self.expires_at = expires_at

    def token_expired(self):
        """
        Return True if the JWT token has expired, False if not.
        """
        return datetime.utcnow() > self.expires_at

    @staticmethod
    def refresh_token(func):
        """
        Use this method decorator to ensure the JWT token is refreshed when needed.
        """
        @wraps(func)
        def inner(self, *args, **kwargs):
            """
            Before calling the wrapped function, we check if the JWT token is expired, and if so, re-connect.
            """
            if self.token_expired():
                self.connect()
            return func(self, *args, **kwargs)
        return inner

    def _load_data(
            self,
            resource,
            detail_resource=None,
            resource_id=None,
            querystring=None,
            should_traverse_pagination=False,
            default=DEFAULT_VALUE_SAFEGUARD,
    ):
        """
        Loads a response from a call to one of the API endpoints.

        Arguments:
            resource: The endpoint resource name.
            detail_resource: The sub-resource to append to the path.
            resource_id: The resource ID for the specific detail to get from the endpoint.
            querystring: Optional query string parameters.
            should_traverse_pagination: Whether to traverse pagination or return paginated response.
            default: The default value to return in case of no response content.

        Returns
            (JSON): Data returned by the API.
        """
        default_val = default if default is not self.DEFAULT_VALUE_SAFEGUARD else {}
        querystring = querystring or {}

        endpoint = getattr(self.client, resource)
        endpoint = getattr(self.client, resource)(resource_id) if resource_id else endpoint
        endpoint = getattr(endpoint, detail_resource) if detail_resource else endpoint
        response = endpoint.get(**querystring)
        if should_traverse_pagination:
            results = traverse_pagination(response, endpoint)
            response = {
                'count': len(results),
                'next': None,
                'previous': None,
                'results': results,
            }

        return response or default_val


def traverse_pagination(response, endpoint):
    """
    Traverse a paginated API response.

    Extracts and concatenates "results" (list of dict) returned by DRF-powered
    APIs.

    Arguments:
        response (Dict): Current response dict from service API
        endpoint (slumber Resource object): slumber Resource object from edx-rest-api-client

    Returns:
        list of dict.

    """
    results = response.get('results', [])

    next_page = response.get('next')
    while next_page:
        querystring = urllib.parse.parse_qs(urllib.parse.urlparse(next_page).query, True)
        response = endpoint.get(**querystring)
        results += response.get('results', [])
        next_page = response.get('next')

    return results



# -*- coding: utf-8 -*-
"""
Tests for filters in enterprise_data.
"""
from __future__ import absolute_import, unicode_literals

import mock
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from enterprise_data.models import EnterpriseEnrollment
from test_utils import UserFactory


class TestConsentGrantedFilterBackend(APITestCase):
    """
    Test suite for the ``TestConsentGrantedFilterBackend`` filter.
    """
    fixtures = ('enterprise_enrollment', 'enterprise_user', )

    def setUp(self):
        super(TestConsentGrantedFilterBackend, self).setUp()
        self.user = UserFactory(is_staff=True)
        self.client.force_authenticate(user=self.user)
        self.url = reverse('v0:enterprise-enrollments-list',
                           kwargs={'enterprise_id': 'ee5e6b3a-069a-4947-bb8d-d2dbc323396c'})

    @mock.patch('enterprise_data.permissions.EnterpriseApiClient')
    def test_filter_for_list(self, mock_enterprise_api_client):
        """
        Filter users through email/username if requesting user is staff, otherwise based off of request user ID.
        """
        mock_enterprise_api_client.return_value.get_with_access_to.return_value = {
            'uuid': 'ee5e6b3a-069a-4947-bb8d-d2dbc323396c'
        }
        response = self.client.get(self.url)
        assert response.status_code == status.HTTP_200_OK
        data = response.json()

        # Fixture data for enterprise 'ee5e6b3a-069a-4947-bb8d-d2dbc323396c' contains 3 objects but only 3 with consent
        assert EnterpriseEnrollment.objects.filter(enterprise_id='ee5e6b3a-069a-4947-bb8d-d2dbc323396c').count() == 3
        assert len(data['results']) == 2



"""
Enterprise_data url configuration.
"""
from __future__ import absolute_import, unicode_literals

from django.conf.urls import include, url

urlpatterns = [
    url(
        r'^enterprise/api/',
        include('enterprise_data.api.urls'),
        name='enterprise_data_api'
    ),
]



# -*- coding: utf-8 -*-
"""
Tests for clients in enterprise_data.
"""
from edx_rest_api_client.exceptions import HttpClientError
from mock import ANY, Mock, mock
from rest_framework.exceptions import NotFound, ParseError

from django.test import TestCase, override_settings

from enterprise_data.clients import EnterpriseApiClient
from test_utils import UserFactory


class TestEnterpriseApiClient(TestCase):
    """
    Test Enterprise API client used to connect to LMS enterprise endpoints
    """

    def setUp(self):
        self.user = UserFactory()
        self.enterprise_id = '0395b02f-6b29-42ed-9a41-45f3dff8349c'
        self.api_response = {
            'count': 1,
            'results': [{
                'enterprise_name': 'Test Enterprise',
                'enterprise_id': 'test-id'
            }]
        }
        self.mocked_get_endpoint = Mock(return_value=self.api_response)
        super(TestEnterpriseApiClient, self).setUp()

    def mock_client(self):
        """
        Set up a mocked Enterprise API Client. Avoiding doing this in setup so we can test __init__.
        """
        self.client = EnterpriseApiClient('test-token')
        setattr(self.client, 'enterprise-learner', Mock(
            get=self.mocked_get_endpoint
        ))

        setattr(self.client, 'enterprise-customer', Mock(
            with_access_to=Mock(get=self.mocked_get_endpoint)
        ))

    @mock.patch('enterprise_data.clients.EdxRestApiClient.__init__')
    def test_inits_client_with_jwt(self, mock_init):
        self.mock_client()
        mock_init.assert_called_with(ANY, jwt='test-token')

    def test_get_enterprise_learner_returns_results_for_user(self):
        self.mock_client()
        results = self.client.get_enterprise_learner(self.user)
        self.mocked_get_endpoint.assert_called_with(username=self.user.username)
        assert(results == self.api_response['results'][0])

    def test_get_enterprise_learner_raises_exception_on_error(self):
        self.mocked_get_endpoint = Mock(side_effect=HttpClientError)
        self.mock_client()
        with self.assertRaises(HttpClientError):
            _ = self.client.get_enterprise_learner(self.user)

    def test_get_enterprise_learner_returns_none_on_empty_results(self):
        self.mocked_get_endpoint = Mock(return_value={
            'count': 0,
            'results': []
        })
        self.mock_client()
        results = self.client.get_enterprise_learner(self.user)
        self.assertIsNone(results)

    def test_get_enterprise_learner_raises_not_found_on_no_results(self):
        self.mocked_get_endpoint = Mock(return_value={})
        self.mock_client()
        with self.assertRaises(NotFound):
            _ = self.client.get_enterprise_learner(self.user)

    def test_get_enterprise_learner_raises_parse_error_on_multiple_results(self):
        self.mocked_get_endpoint = Mock(return_value={
            'count': 2,
            'results': [{}, {}]
        })
        self.mock_client()
        with self.assertRaises(ParseError):
            _ = self.client.get_enterprise_learner(self.user)

    def test_get_with_access_to_returns_results_for_user(self):
        self.mock_client()
        results = self.client.get_with_access_to(self.user, self.enterprise_id)
        self.mocked_get_endpoint.assert_called_with(
            permissions=[EnterpriseApiClient.ENTERPRISE_DATA_API_GROUP],
            enterprise_id=self.enterprise_id
        )
        assert(results == self.api_response['results'][0])

    def test_get_with_access_to_raises_exception_on_error(self):
        self.mocked_get_endpoint = Mock(side_effect=HttpClientError)
        self.mock_client()
        with self.assertRaises(HttpClientError):
            _ = self.client.get_with_access_to(self.user, self.enterprise_id)

    def test_get_with_access_to_returns_none_on_empty_results(self):
        self.mocked_get_endpoint = Mock(return_value={
            'count': 0,
            'results': []
        })
        self.mock_client()
        results = self.client.get_with_access_to(self.user, self.enterprise_id)
        self.assertIsNone(results)

    def test_get_with_access_to_raises_not_found_on_no_results(self):
        self.mocked_get_endpoint = Mock(return_value={})
        self.mock_client()
        with self.assertRaises(NotFound):
            _ = self.client.get_with_access_to(self.user, self.enterprise_id)

    def test_get_with_access_to_raises_parse_error_on_multiple_results(self):
        self.mocked_get_endpoint = Mock(return_value={
            'count': 2,
            'results': [{}, {}]
        })
        self.mock_client()
        with self.assertRaises(ParseError):
            _ = self.client.get_with_access_to(self.user, self.enterprise_id)



"""
These settings are here to use during tests, because django requires them.

In a real-world use case, apps in this project are installed into other
Django applications, so these settings will not be used.
"""
from __future__ import absolute_import, unicode_literals

from os.path import abspath, dirname, join


def here(*args):
    """
    Return the absolute path to a directory from this file.
    """
    return join(abspath(dirname(__file__)), *args)


def root(*args):
    """
    Return the absolute path to some file from the project's root.
    """
    return abspath(join(abspath(here('../..')), *args))


DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "default.db",
        "USER": "",
        "PASSWORD": "",
        "HOST": "",
        "PORT": "",
    }
}

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sites",
    "django.contrib.sessions",
    "django.contrib.admin",  # only used in DEBUG mode
    "django.contrib.messages",
    "django.contrib.staticfiles",

    "enterprise_data",
    "enterprise_reporting"
)

MIDDLEWARE_CLASSES = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]

MIDDLEWARE = MIDDLEWARE_CLASSES  # Django 1.10 compatibility - the setting was renamed

AUTHENTICATION_BACKENDS = ["django.contrib.auth.backends.ModelBackend"]

SESSION_ENGINE = "django.contrib.sessions.backends.file"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": {
                "django.contrib.auth.context_processors.auth",  # this is required for admin
                "django.contrib.messages.context_processors.messages",
            }
        }
    },
]

LMS_BASE_URL = "http://localhost:8000/"

ROOT_URLCONF = "enterprise_data.urls"

SECRET_KEY = "insecure-secret-key"

USE_TZ = True
TIME_ZONE = 'UTC'



# -*- coding: utf-8 -*-
"""
Classes that handle sending reports for enterprise customers with specific delivery methods.
"""

from __future__ import absolute_import, unicode_literals

import logging
import os
from smtplib import SMTPException

import paramiko

from enterprise_reporting.utils import compress_and_encrypt, send_email_with_attachment

LOGGER = logging.getLogger(__name__)


class DeliveryMethod(object):
    """
    Base class for different types of Enterprise report delivery methods.
    """

    def __init__(self, reporting_config, password):
        """Initialize the SFTP Delivery Method."""
        self.enterprise_customer_name = reporting_config['enterprise_customer']['name']
        self.data_type = reporting_config['data_type']
        self.report_type = reporting_config['report_type']
        self.password = password

    def send(self, files):
        """Base method for sending files, to perform common sending logic."""
        LOGGER.info('Encrypting data report for {}'.format(self.enterprise_customer_name))
        return compress_and_encrypt(files, self.password)


class SMTPDeliveryMethod(DeliveryMethod):
    """
    Class that handles sending an enterprise report file via email as an encrypted zip file.
    """

    REPORT_EMAIL_FROM_EMAIL = os.environ.get('SEND_EMAIL_FROM')
    REPORT_EMAIL_SUBJECT = '{enterprise_name} edX Learner Data'
    REPORT_EMAIL_BODY = """
    Please find the attached {type} data for courses on edX.
    For any questions or concerns, please contact your edX sales representative.
    Thanks,
    The edX for Business Team
    """

    def __init__(self, reporting_config, password):
        """Initialize the SMTP Delivery Method."""
        super().__init__(reporting_config, password)
        self._email = reporting_config['email']

    @property
    def email(self):
        return self._email

    @email.setter
    def email(self, value):
        """Make the internal email value a list if there are multiple emails."""
        self._email = value if isinstance(value, list) else [value]

    def send(self, files):
        """Send the given files in zip format through SMTP."""
        data_report_zipped = super().send(files)
        LOGGER.info('Emailing encrypted data as a ZIP to {}'.format(self.enterprise_customer_name))
        try:
            send_email_with_attachment(
                self.REPORT_EMAIL_SUBJECT.format(enterprise_name=self.enterprise_customer_name),
                self.REPORT_EMAIL_BODY.format(type=self.data_type),
                self.REPORT_EMAIL_FROM_EMAIL,
                self.email,
                data_report_zipped
            )
        except SMTPException:
            LOGGER.exception('Failed to send email report to {} for {}'.format(
                self.email,
                self.enterprise_customer_name
            ))
        else:
            LOGGER.info('Email report with encrypted zip successfully sent to {} for {}'.format(
                self.email,
                self.enterprise_customer_name
            ))


class SFTPDeliveryMethod(DeliveryMethod):
    """
    Class that handles sending an enterprise report file via SFTP.
    """

    def __init__(self, reporting_config, password):
        """Initialize the SFTP Delivery Method."""
        super().__init__(reporting_config, password)
        self.hostname = reporting_config['sftp_hostname']
        self.port = reporting_config['sftp_port']
        self.username = reporting_config['sftp_username']
        self.file_path = reporting_config['sftp_file_path']

    def send(self, files):
        """Send the given files in zip format through SFTP."""
        data_report_zipped = super().send(files)
        LOGGER.info('Connecting via SFTP to remote host {} for {}'.format(
            self.hostname,
            self.enterprise_customer_name
        ))
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(
            hostname=self.hostname,
            port=self.port,
            username=self.username,
            password=self.password,
        )
        sftp = ssh.open_sftp()
        sftp.put(
            data_report_zipped,
            os.path.join(self.file_path, os.path.basename(data_report_zipped))
        )
        sftp.close()
        ssh.close()
        LOGGER.info('Successfully sent report via sftp for {}'.format(self.enterprise_customer_name))



#!/usr/bin/python3
"""
Export edx data (mysql or vertica) into a csv file.

This script provides the framework for querying the edxapp mysql database
or the Vertica-based data warehouse and outputting the result of your
query to a CSV file.

This is useful for analyzing data contained in these data sources
and sharing your analysis with others.

To run this script:

1. Create edx-enterprise-data virtualenv.
2. Install python requirements with `pip install -r requirements/dev.txt`.
3. Set environment variables required by mysql-connector:
    $ export MYSQLXPB_PROTOBUF_INCLUDE_DIR=/usr/local/opt/protobuf/include
    $ export MYSQLXPB_PROTOBUF_LIB_DIR=/usr/local/opt/protobuf/lib
    $ export MYSQLXPB_PROTOC=/usr/local/opt/protobuf/bin/protoc
4. Set environment variables required to connect to mysql:
    $ export MYSQL_HOST=[mysql hostname]
    $ export MYSQL_USERNAME=[mysql username]
    $ export MYSQL_PASSWORD=[mysql password]
    $ export MYSQL_DATABASE=[mysql database name]
5. Run `python enterprise_reporting/export_edxapp_data.py`.
6. Find CSV output in the .output directory at the root of this repository.
"""
from __future__ import absolute_import, unicode_literals

import csv
import datetime
import logging
import os
import sys
from io import open  # pylint: disable=redefined-builtin

import mysql.connector
import vertica_python

logging.basicConfig(stream=sys.stdout, level=logging.INFO)

MISSING_ENROLLMENT_QUERY = '''
    SELECT
        u.id,
        u.username,
        u.email,
        ec.name AS enterprise_name,
        sce.course_id,
        sce.mode,
        sce.created AS course_enrollment_created,
        ecu.created AS enterprise_link_created,
        u.date_joined AS edx_account_created
    FROM
        enterprise_enterprisecustomeruser ecu
    JOIN
        enterprise_enterprisecustomer ec
    ON
        ecu.enterprise_customer_id = ec.uuid
    JOIN
        auth_user u
    ON
        ecu.user_id = u.id AND
        u.email NOT LIKE '%@edx.org'
    JOIN
        student_courseenrollment sce
    ON
        ecu.user_id = sce.user_id AND
        ecu.created <= sce.created
    LEFT JOIN
        enterprise_enterprisecourseenrollment ece
    ON
        ecu.id = ece.enterprise_customer_user_id
    WHERE
        ece.id IS NULL
    ORDER BY
        ec.name,
        sce.created DESC;
'''


def export_data(data, report_name_prefix):
    """
    Export given data to a csv report.
    """
    if data:
        now = datetime.datetime.now()
        output_dir = '.output'
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        csv_filename = '{}/{}_{}.csv'.format(
            output_dir,
            report_name_prefix,
            now.strftime('%Y%m%d%H%M')
        )
        header = data[0].keys()
        with open(csv_filename, 'w', newline='') as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(header)
            for row in data:
                for key, value in row.items():
                    if isinstance(value, datetime.datetime):
                        row[key] = value.strftime('%Y-%m-%d %H:%M:%S')
                    else:
                        row[key] = str(value)
                writer.writerow(row.values())


def fetch_lms_data(query):
    """
    Return the results of the given query against a mysql database.

    Uses the following environment variables to connnect to the database:
    MYSQL_HOST, MYSQL_HOST, MYSQL_PASSWORD, MYSQL_DATABASE
    """
    data = []
    connection_info = {
        'host': os.environ.get('MYSQL_HOST'),
        'user': os.environ.get('MYSQL_USERNAME'),
        'password': os.environ.get('MYSQL_PASSWORD'),
        'database': os.environ.get('MYSQL_DATABASE'),
    }
    try:
        connection = mysql.connector.connect(**connection_info)
        cur = connection.cursor(dictionary=True)
        cur.execute(query)
        for row in cur:
            data.append(row)
    finally:
        connection.close()
    return data


def fetch_vertica_data(query):
    """
    Return the results of the given query against a vertica database.

    Uses the following environment variables to connnect to the database:
    VERTICA_HOST, VERTICA_USERNAME, VERTICA_PASSWORD, VERTICA_DATABASE
    """
    connection_info = {
        'host': os.environ.get('VERTICA_HOST'),
        'user': os.environ.get('VERTICA_USERNAME'),
        'password': os.environ.get('VERTICA_PASSWORD'),
        'database': os.environ.get('VERTICA_DATABASE'),
    }
    with vertica_python.connect(**connection_info) as connection:
        cur = connection.cursor('dict')
        cur.execute(query)
        return cur.fetchall()


def export_missing_enterprise_enrollments():
    """
    Find and export missing enterprise enrollments.

    Run query to find course enrollments which could potentially be
    considered Enterprise enrollments, but are not currently recorded
    as such.
    """
    missing_enterprise_enrollments = fetch_lms_data(MISSING_ENROLLMENT_QUERY)
    if missing_enterprise_enrollments:
        export_data(missing_enterprise_enrollments, 'missing_enterprise_enrollments')


if __name__ == "__main__":
    export_missing_enterprise_enrollments()
    sys.exit(0)



# -*- coding: utf-8 -*-
"""
Database models for enterprise data.
"""
from __future__ import absolute_import, unicode_literals

from logging import getLogger

from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

LOGGER = getLogger(__name__)


@python_2_unicode_compatible
class EnterpriseEnrollment(models.Model):
    """Enterprise Enrollment is the learner details for a specific course enrollment.

    This information includes a mix of the learners meta data (email, username, etc), the enterprise they are
    associated with (enterprise_id, enterprise_name, etc), and their status in the course (course_id,
    letter_grade, has_passed, etc).
    This is meant to closely mirror the data the Enterprises recieve in the automated report (see enterprise_reporting
    folder in this repo), with the exception of some data that is only calculated in Vertica as this data is pulled from
    the analytics result store via the analytics-data-api project.
    """

    class Meta:
        app_label = 'enterprise_data'
        db_table = 'enterprise_enrollment'
        verbose_name = _("Enterprise Enrollment")
        verbose_name_plural = _("Enterprise Enrollments")

    enterprise_id = models.UUIDField()
    enterprise_name = models.CharField(max_length=255)
    lms_user_id = models.PositiveIntegerField()
    enterprise_user = models.ForeignKey(
        'EnterpriseUser',
        related_name='enrollments',
        to_field='enterprise_user_id',
    )
    course_id = models.CharField(max_length=255, help_text='The course the learner is enrolled in.')
    enrollment_created_timestamp = models.DateTimeField()
    user_current_enrollment_mode = models.CharField(max_length=32)
    consent_granted = models.NullBooleanField(default=None)
    letter_grade = models.CharField(max_length=32, null=True)
    has_passed = models.BooleanField(default=False)
    passed_timestamp = models.DateTimeField(null=True)
    enterprise_sso_uid = models.CharField(max_length=255, null=True)
    enterprise_site_id = models.PositiveIntegerField(null=True)
    course_title = models.CharField(max_length=255, null=True)
    course_start = models.DateTimeField(null=True)
    course_end = models.DateTimeField(null=True)
    course_pacing_type = models.CharField(max_length=32, null=True)
    course_duration_weeks = models.CharField(max_length=32, null=True)
    course_min_effort = models.PositiveIntegerField(null=True)
    course_max_effort = models.PositiveIntegerField(null=True)
    user_account_creation_timestamp = models.DateTimeField(null=True)
    user_email = models.CharField(max_length=255, null=True)
    user_username = models.CharField(max_length=255, null=True)
    course_key = models.CharField(max_length=255, null=True)
    user_country_code = models.CharField(max_length=2, null=True)
    last_activity_date = models.DateField(null=True)
    coupon_name = models.CharField(max_length=128, null=True)
    coupon_code = models.CharField(max_length=128, null=True)
    offer = models.CharField(max_length=128, null=True)
    current_grade = models.FloatField(null=True)
    course_price = models.DecimalField(decimal_places=2, max_digits=12, null=True)
    discount_price = models.DecimalField(decimal_places=2, max_digits=12, null=True)
    created = models.DateTimeField(null=True)

    def __str__(self):
        """
        Return a human-readable string representation of the object.
        """
        return "<Enterprise Enrollment for user {user} in {course}>".format(
            user=self.enterprise_user_id,
            course=self.course_id
        )

    def __repr__(self):
        """
        Return uniquely identifying string representation.
        """
        return self.__str__()


@python_2_unicode_compatible
class EnterpriseUser(models.Model):
    """Information includes a mix of the user's meta data.

    Includes (lms_user_id, username, etc), and the enterprise they are associated with (enterprise_id).
    """

    class Meta:
        app_label = 'enterprise_data'
        db_table = 'enterprise_user'
        verbose_name = _("Enterprise User")
        verbose_name_plural = _("Enterprise Users")
        ordering = ['-user_email']

    enterprise_id = models.UUIDField()
    lms_user_id = models.PositiveIntegerField()
    enterprise_user_id = models.PositiveIntegerField(unique=True)
    enterprise_sso_uid = models.CharField(max_length=255, null=True)
    user_account_creation_timestamp = models.DateTimeField(null=True)
    user_email = models.CharField(max_length=255, null=True)
    user_username = models.CharField(max_length=255, null=True)
    user_country_code = models.CharField(max_length=2, null=True)
    last_activity_date = models.DateField(null=True)
    created = models.DateTimeField(null=True)

    def __str__(self):
        """
        Return a human-readable string representation of the object.
        """
        return "<Enterprise User {user} in {enterprise}>".format(
            user=self.lms_user_id,
            enterprise=self.enterprise_id
        )

    def __repr__(self):
        """
        Return uniquely identifying string representation.
        """
        return self.__str__()



# -*- coding: utf-8 -*-
"""
Views for enterprise api version 0 endpoint.
"""
from __future__ import absolute_import, unicode_literals

from datetime import date, timedelta
from logging import getLogger

from edx_rest_framework_extensions.authentication import JwtAuthentication
from edx_rest_framework_extensions.paginators import DefaultPagination
from rest_framework import filters, viewsets
from rest_framework.decorators import list_route
from rest_framework.exceptions import NotFound
from rest_framework.response import Response

from django.db.models import Count, Max
from django.utils import timezone

from enterprise_data.api.v0 import serializers
from enterprise_data.filters import CONSENT_TRUE_OR_NONE_Q, ConsentGrantedFilterBackend
from enterprise_data.models import EnterpriseEnrollment, EnterpriseUser
from enterprise_data.permissions import HasDataAPIDjangoGroupAccess

LOGGER = getLogger(__name__)


def subtract_one_month(original_date):
    """
    Returns a date exactly one month prior to the passed in date.
    """
    one_day = timedelta(days=1)
    one_month_earlier = original_date - one_day
    while one_month_earlier.month == original_date.month or one_month_earlier.day > original_date.day:
        one_month_earlier -= one_day
    return one_month_earlier


class EnterpriseViewSet(viewsets.ViewSet):
    """
    Base class for all Enterprise view sets.
    """
    authentication_classes = (JwtAuthentication,)
    pagination_class = DefaultPagination
    permission_classes = (HasDataAPIDjangoGroupAccess,)

    def ensure_data_exists(self, request, data, error_message=None):
        """
        Ensure that the API response brings us valid data. If not, raise an error and log it.
        """
        if not data:
            error_message = (
                error_message or "Unable to fetch API response from endpoint '{}'.".format(request.get_full_path())
            )
            LOGGER.error(error_message)
            raise NotFound(error_message)

    def paginate_queryset(self, queryset):
        """
        Allows no_page query param to skip pagination
        """
        if 'no_page' in self.request.query_params:
            return None
        return super(EnterpriseViewSet, self).paginate_queryset(queryset)


class EnterpriseEnrollmentsViewSet(EnterpriseViewSet, viewsets.ModelViewSet):
    """
    Viewset for routes related to Enterprise course enrollments.
    """
    serializer_class = serializers.EnterpriseEnrollmentSerializer
    filter_backends = (ConsentGrantedFilterBackend, filters.OrderingFilter,)
    ordering_fields = '__all__'
    ordering = ('user_email',)
    CONSENT_GRANTED_FILTER = 'consent_granted'

    def get_queryset(self):
        """
        Returns all learner enrollment records for a given enterprise.
        """
        enterprise_id = self.kwargs['enterprise_id']
        enrollments = EnterpriseEnrollment.objects.filter(enterprise_id=enterprise_id)

        enrollments = self.apply_filters(enrollments)

        self.ensure_data_exists(
            self.request,
            enrollments,
            error_message=(
                "No course enrollments are associated with Enterprise {enterprise_id} from endpoint '{path}'."
                .format(
                    enterprise_id=enterprise_id,
                    path=self.request.get_full_path()
                )
            )
        )
        return enrollments

    def apply_filters(self, queryset):
        """
        Filters enrollments based on query params.
        """
        query_filters = self.request.query_params

        past_week_date = date.today() - timedelta(weeks=1)
        past_month_date = subtract_one_month(date.today())

        passed_date_param = query_filters.get('passed_date')
        if passed_date_param == 'last_week':
            queryset = self.filter_past_week_completions(queryset)

        learner_activity_param = query_filters.get('learner_activity')
        if learner_activity_param == 'active_past_week':
            queryset = self.filter_active_learners(queryset, past_week_date)
        elif learner_activity_param == 'inactive_past_week':
            queryset = self.filter_inactive_learners(queryset, past_week_date)
        elif learner_activity_param == 'inactive_past_month':
            queryset = self.filter_inactive_learners(queryset, past_month_date)

        return queryset

    def filter_distinct_learners(self, queryset):
        """
        Filters queryset to include enrollments with a distinct `enterprise_user_id`.
        """
        return queryset.values_list('enterprise_user_id', flat=True).distinct()

    def filter_active_learners(self, queryset, last_activity_date):
        """
        Filters queryset to include enrollments more recent than the specified `last_activity_date`.
        """
        return queryset.filter(last_activity_date__gte=last_activity_date)

    def filter_inactive_learners(self, queryset, last_activity_date):
        """
        Filters queryset to include enrollments more older than the specified `last_activity_date`.
        """
        return queryset.filter(last_activity_date__lte=last_activity_date)

    def filter_distinct_active_learners(self, queryset, last_activity_date):
        """
        Filters queryset to include distinct enrollments more recent than the specified `last_activity_date`.
        """
        return self.filter_distinct_learners(self.filter_active_learners(queryset, last_activity_date))

    def filter_course_completions(self, queryset):
        """
        Filters queryset to include only enrollments that are passing (i.e., completion).
        """
        return queryset.filter(has_passed=1)

    def filter_past_week_completions(self, queryset):
        """
        Filters only those learners who completed a course in last week.
        """
        date_today = date.today()
        date_week_before = date_today - timedelta(weeks=1)
        return queryset.filter(
            has_passed=True,
            passed_timestamp__lte=date_today,
            passed_timestamp__gte=date_week_before,
        )

    def filter_number_of_users(self):
        """
        Returns number of enterprise users (enrolled AND not enrolled learners)
        """
        enterprise_id = self.kwargs['enterprise_id']
        return EnterpriseUser.objects.filter(enterprise_id=enterprise_id)

    def get_max_created_date(self, queryset):
        """
        Gets the maximum created timestamp from the queryset.
        """
        created_max = queryset.aggregate(Max('created'))
        return created_max['created__max']

    @list_route()
    def overview(self, request, **kwargs):  # pylint: disable=unused-argument
        """
        Returns the following data:
            - # of enrolled learners;
            - # of active learners in the past week/month;
            - # of course completions.
            - # of enterprise users (LMS)
        """
        enrollments = self.get_queryset()
        enrollments = self.filter_queryset(enrollments)
        course_completions = self.filter_course_completions(enrollments)
        distinct_learners = self.filter_distinct_learners(enrollments)
        past_week_date = date.today() - timedelta(weeks=1)
        past_month_date = subtract_one_month(date.today())
        active_learners_week = self.filter_distinct_active_learners(enrollments, past_week_date)
        last_updated_date = self.get_max_created_date(enrollments)
        active_learners_month = self.filter_distinct_active_learners(enrollments, past_month_date)
        enterprise_users = self.filter_number_of_users()
        content = {
            'enrolled_learners': distinct_learners.count(),
            'active_learners': {
                'past_week': active_learners_week.count(),
                'past_month': active_learners_month.count(),
            },
            'course_completions': course_completions.count(),
            'last_updated_date': last_updated_date,
            'number_of_users': enterprise_users.count(),
        }
        return Response(content)


class EnterpriseUsersViewSet(EnterpriseViewSet, viewsets.ModelViewSet):
    """
    Viewset for routes related to Enterprise users.
    """
    queryset = EnterpriseUser.objects.all()
    serializer_class = serializers.EnterpriseUserSerializer
    filter_backends = (filters.OrderingFilter,)
    ordering_fields = '__all__'
    ordering = ('user_email',)

    def list(self, request, **kwargs):
        """
        List view for learner records for a given enterprise.
        """
        enterprise_id = kwargs['enterprise_id']
        users = self.queryset.filter(enterprise_id=enterprise_id)
        # Ignore users that only have declined consent on all enrollments
        users = users.filter(CONSENT_TRUE_OR_NONE_Q).distinct()

        has_enrollments = request.query_params.get('has_enrollments')
        if has_enrollments == 'true':
            users = users.filter(CONSENT_TRUE_OR_NONE_Q, enrollments__isnull=False).distinct()
        elif has_enrollments == 'false':
            users = users.filter(CONSENT_TRUE_OR_NONE_Q, enrollments__isnull=True)

        active_courses = request.query_params.get('active_courses')
        if active_courses == 'true':
            users = users.filter(CONSENT_TRUE_OR_NONE_Q, enrollments__course_end__gte=timezone.now())
        elif active_courses == 'false':
            users = users.filter(CONSENT_TRUE_OR_NONE_Q, enrollments__course_end__lte=timezone.now())

        all_enrollments_passed = request.query_params.get('all_enrollments_passed')
        if all_enrollments_passed == 'true':
            users = users.exclude(enrollments__has_passed=False)
        elif all_enrollments_passed == 'false':
            users = users.filter(enrollments__has_passed=False)

        # do the sorting
        users = self.filter_queryset(users)

        # Bit to account for pagination
        page = self.paginate_queryset(users)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)
        serializer = self.get_serializer(users, many=True)
        return Response(serializer.data)


class EnterpriseLearnerCompletedCoursesViewSet(EnterpriseViewSet, viewsets.ModelViewSet):
    """
    View to manage enterprise learner completed course enrollments.
    """
    serializer_class = serializers.LearnerCompletedCoursesSerializer
    filter_backends = (filters.OrderingFilter,)
    ordering_fields = '__all__'
    ordering = ('user_email',)

    def get_queryset(self):
        """
        Returns number of completed courses against each learner.
        """
        enterprise_id = self.kwargs['enterprise_id']
        # Get the number of completed courses against a learner.
        enrollments = EnterpriseEnrollment.objects.filter(
            enterprise_id=enterprise_id,
            has_passed=True,
            consent_granted=True,
        ).values('user_email').annotate(completed_courses=Count('course_id')).order_by('user_email')
        return enrollments



# -*- coding: utf-8 -*-
"""
Test factories.
"""

from datetime import datetime

import factory
from faker import Factory as FakerFactory

from django.contrib.auth.models import User

from enterprise_data.models import EnterpriseEnrollment, EnterpriseUser

FAKER = FakerFactory.create()


class EnterpriseEnrollmentFactory(factory.django.DjangoModelFactory):
    """
    EnterpriseCourseEnrollment factory.

    Creates an instance of EnterpriseCourseEnrollment with minimal boilerplate.
    """

    class Meta(object):
        """
        Meta for EnterpriseCourseEnrollmentFactory.
        """

        model = EnterpriseEnrollment

    id = factory.lazy_attribute(lambda x: FAKER.random_int(min=1))
    enterprise_id = factory.lazy_attribute(lambda x: 'ee5e6b3a069a4947bb8dd2dbc323396c')
    lms_user_id = factory.lazy_attribute(lambda x: FAKER.random_int(min=1))
    course_id = factory.lazy_attribute(lambda x: FAKER.slug())
    enrollment_created_timestamp = factory.lazy_attribute(lambda x: '2018-01-01')
    user_current_enrollment_mode = factory.lazy_attribute(lambda x: 'verified')
    has_passed = False


class UserFactory(factory.django.DjangoModelFactory):
    """
    User Factory.

    Creates an instance of User with minimal boilerplate
    """
    class Meta(object):
        model = User
        django_get_or_create = ('email', 'username')

    _DEFAULT_PASSWORD = 'test'

    username = factory.Sequence(u'robot{0}'.format)
    email = factory.Sequence(u'robot+test+{0}@edx.org'.format)
    password = factory.PostGenerationMethodCall('set_password', _DEFAULT_PASSWORD)
    first_name = factory.Sequence(u'Robot{0}'.format)
    last_name = 'Test'
    is_staff = factory.lazy_attribute(lambda x: False)
    is_active = True
    is_superuser = False
    last_login = datetime(2012, 1, 1)
    date_joined = datetime(2011, 1, 1)


class EnterpriseUserFactory(factory.django.DjangoModelFactory):
    """
    Enterprise User Factory.

    Creates an instance of Enteprise User with minimal boilerplate
    """
    class Meta(object):
        model = EnterpriseUser

    enterprise_id = factory.lazy_attribute(lambda x: 'ee5e6b3a069a4947bb8dd2dbc323396c')
    lms_user_id = factory.lazy_attribute(lambda x: FAKER.random_int(min=1))
    enterprise_user_id = factory.lazy_attribute(lambda x: FAKER.random_int(min=1))
    enterprise_sso_uid = factory.lazy_attribute(lambda x: FAKER.text(max_nb_chars=255))
    user_account_creation_timestamp = datetime(2011, 1, 1)
    user_username = factory.Sequence(u'robot{0}'.format)
    user_email = factory.Sequence(u'robot+test+{0}@edx.org'.format)
    user_country_code = factory.lazy_attribute(lambda x: FAKER.country_code())
    last_activity_date = datetime(2012, 1, 1).date()



"""
Tests for permissions
"""

from logging import getLogger

from mock import MagicMock, patch
from pytest import mark

from django.test import TestCase

from enterprise_data.permissions import HasDataAPIDjangoGroupAccess, IsStaffOrEnterpriseUser
from test_utils import UserFactory

LOGGER = getLogger(__name__)


class PermissionsTestCase(TestCase):
    """
    Common logic for permissions class tests.
    """
    def setUp(self):
        super(PermissionsTestCase, self).setUp()
        self.enterprise_id = 'fake-enterprise-id'
        self.user = UserFactory()
        self.request = MagicMock(
            user=self.user,
            parser_context={
                'kwargs': {
                    'enterprise_id': self.enterprise_id
                }
            },
            session={},
            auth=MagicMock()
        )

        enterprise_api_client = patch('enterprise_data.permissions.EnterpriseApiClient')
        self.enterprise_api_client = enterprise_api_client.start()
        self.addCleanup(enterprise_api_client.stop)


@mark.django_db
class TestIsStaffOrEnterpriseUser(PermissionsTestCase):
    """
    Tests of the IsStaffOrEnterpriseUser permission
    """

    def setUp(self):
        super(TestIsStaffOrEnterpriseUser, self).setUp()
        self.permission = IsStaffOrEnterpriseUser()

    def test_staff_access(self):
        self.user.is_staff = True
        self.assertTrue(self.permission.has_permission(self.request, None))

    def test_enterprise_learner_has_access(self):
        self.enterprise_api_client.return_value.get_enterprise_learner.return_value = {
            'enterprise_customer': {
                'uuid': self.enterprise_id
            },
            'groups': ['enterprise_data_api_access'],
        }
        self.assertTrue(self.permission.has_permission(self.request, None))

    def test_wrong_enterprise_learner_has_no_access(self):
        self.enterprise_api_client.return_value.get_enterprise_learner.return_value = {
            'enterprise_customer': {
                'uuid': 'some-other-enterprise-id'
            },
            'groups': ['enterprise_data_api_access'],
        }
        self.assertFalse(self.permission.has_permission(self.request, None))

    def test_enterprise_learner_has_no_group_access(self):
        self.enterprise_api_client.return_value.get_enterprise_learner.return_value = {
            'enterprise_customer': {
                'uuid': 'some-other-enterprise-id'
            },
            'groups': [],
        }
        self.assertFalse(self.permission.has_permission(self.request, None))

    def test_no_enterprise_learner_for_user(self):
        self.enterprise_api_client.return_value.get_enterprise_learner.return_value = {}
        self.assertFalse(self.permission.has_permission(self.request, None))


@mark.django_db
class TestHasDataAPIDjangoGroupAccess(PermissionsTestCase):
    """
    Tests of the IsStaffOrEnterpriseUser permission
    """

    def setUp(self):
        super(TestHasDataAPIDjangoGroupAccess, self).setUp()
        self.permission = HasDataAPIDjangoGroupAccess()

    def test_staff_access_without_group_permission(self):
        self.user.is_staff = True
        self.enterprise_api_client.return_value.get_with_access_to.return_value = {}
        self.assertFalse(self.permission.has_permission(self.request, None))

    def test_staff_access_with_group_permission(self):
        self.user.is_staff = True
        self.enterprise_api_client.return_value.get_with_access_to.return_value = {
            'uuid': self.enterprise_id
        }
        self.assertTrue(self.permission.has_permission(self.request, None))

    def test_enterprise_user_has_access_with_group_permission(self):
        self.enterprise_api_client.return_value.get_with_access_to.return_value = {
            'uuid': self.enterprise_id
        }
        self.assertTrue(self.permission.has_permission(self.request, None))

    def test_enterprise_user_without_group_permission(self):
        self.enterprise_api_client.return_value.get_with_access_to.return_value = {}
        self.assertFalse(self.permission.has_permission(self.request, None))



"""
Clients used to connect to other systems.
"""
from __future__ import absolute_import, unicode_literals

import logging

from edx_rest_api_client.client import EdxRestApiClient
from edx_rest_api_client.exceptions import HttpClientError, HttpServerError
from rest_framework.exceptions import NotFound, ParseError

from django.conf import settings

LOGGER = logging.getLogger('enterprise_data')


class EnterpriseApiClient(EdxRestApiClient):
    """
    The EnterpriseApiClient is used to communicate with the enterprise API endpoints in the LMS.

    This class is a sub-class of the edX Rest API Client
    (https://github.com/edx/edx-rest-api-client).
    """

    API_BASE_URL = settings.LMS_BASE_URL + 'enterprise/api/v1/'
    ENTERPRISE_DATA_API_GROUP = 'enterprise_data_api_access'

    def __init__(self, jwt):
        """
        Initialize client with given jwt.
        """
        super(EnterpriseApiClient, self).__init__(self.API_BASE_URL, jwt=jwt)

    def get_enterprise_learner(self, user):
        """
        Get an enterprise learner record for a given user with the enterprise association.

        Returns: learner record or None if unable to retrieve or no Enterprise Learner exists
        """
        try:
            querystring = {'username': user.username}
            endpoint = getattr(self, 'enterprise-learner')
            response = endpoint.get(**querystring)
        except (HttpClientError, HttpServerError) as exc:
            LOGGER.warning("Unable to retrieve Enterprise Customer Learner details for user {}: {}"
                           .format(user.username, exc))
            raise exc

        if response.get('results', None) is None:
            raise NotFound('Unable to process Enterprise Customer Learner details for user {}: No Results Found'
                           .format(user.username))

        if response['count'] > 1:
            raise ParseError('Multiple Enterprise Customer Learners found for user {}'.format(user.username))

        if response['count'] == 0:
            return None

        return response['results'][0]

    def get_with_access_to(self, user, enterprise_id):
        """
        Get the enterprises that this user has access to for the data api permission django group.
        """
        try:
            querystring = {
                'permissions': [self.ENTERPRISE_DATA_API_GROUP],
                'enterprise_id': enterprise_id,
            }
            endpoint = getattr(self, 'enterprise-customer')
            endpoint = getattr(endpoint, 'with_access_to')
            response = endpoint.get(**querystring)
        except (HttpClientError, HttpServerError) as exc:
            LOGGER.warning("Unable to retrieve Enterprise Customer with_access_to details for user {}: {}"
                           .format(user.username, exc))
            raise exc

        if response.get('results', None) is None:
            raise NotFound('Unable to process Enterprise Customer with_access_to details for user {}, enterprise {}:'
                           ' No Results Found'
                           .format(user.username, enterprise_id))

        if response['count'] > 1:
            raise ParseError('Multiple Enterprise Customers found for user {}, enterprise id {}'
                             .format(user.username, enterprise_id))

        if response['count'] == 0:
            return None

        return response['results'][0]



# -*- coding: utf-8 -*-
"""
Tests for the `enterprise-data` models module.
"""
import unittest

import ddt
from pytest import mark

from test_utils import EnterpriseEnrollmentFactory, EnterpriseUserFactory


@mark.django_db
@ddt.ddt
class TestEnterpriseEnrollment(unittest.TestCase):
    """
    Tests for Enterprise Enrollment model
    """

    def setUp(self):
        enterprise_user = EnterpriseUserFactory(enterprise_user_id=1234)
        course_id = 'course-v1:edX+DemoX+DemoCourse'

        self.enrollment = EnterpriseEnrollmentFactory(
            enterprise_user=enterprise_user,
            course_id=course_id
        )
        super(TestEnterpriseEnrollment, self).setUp()

    @ddt.data(str, repr)
    def test_string_conversion(self, method):
        """
        Test conversion to string.
        """
        expected_str = '<Enterprise Enrollment for user 1234 in course-v1:edX+DemoX+DemoCourse>'
        assert expected_str == method(self.enrollment)


@mark.django_db
@ddt.ddt
class TestEnterpriseUser(unittest.TestCase):
    """
    Tests for Enterprise User model
    """
    def setUp(self):
        enterprise_id = 'ee5e6b3a-069a-4947-bb8d-d2dbc323396c'
        lms_user_id = 1234
        self.user = EnterpriseUserFactory(lms_user_id=lms_user_id, enterprise_id=enterprise_id)
        super(TestEnterpriseUser, self).setUp()

    @ddt.data(str, repr)
    def test_string_conversion(self, method):
        """
        Test conversion to string.
        """
        expected_str = '<Enterprise User 1234 in ee5e6b3a-069a-4947-bb8d-d2dbc323396c>'
        assert expected_str == method(self.user)
